######## Includes all functions used to find determinantal varieties, suitable subsystems of them and compute their mixed volume

##1. graph2CM

# INPUT: minimally rigid graph in dimension dim
# OUTPUT: symbolic  CM matrix, coefficients of CM- c, variables of CM- x, ring of matrix- R 

def graph2CM(G,dim):
	
	m=G.order()
	n=G.size()
	A=G.adjacency_matrix()
	#L=3*m-binomial(dim+1,2)
	l=(m*(m-1)-2*n)/2
	
	"Set Variables"
	x=list(var('x_%d' % (i+1)) for i in range(l))
	
	"Set distances"
	c=list(var(','.join(['c_%d' % (i+1) for i in range(n)])))
	R=PolynomialRing(QQ, x+c)
	All_vars=R(0).args()
	x=All_vars[0:l]
	c=All_vars[l:l+n]
	"D: variable-coefficients pre-CM matrix"
	D=matrix(R,m)
	
	"variable counter"
	k1=0 
	
	"distance counter"
	k2=0 

	for i in range(m-1):
		for j in range(i+1,m):
			if A[i,j]==1:
				D[i,j]=c[k2]
				k2=k2+1
			else:
				D[i,j]=x[k1]
				k1=k1+1
			D[j,i]=D[i,j]
			
	
	
	
	CM=matrix(R,m+1)
	CM[1:,1:]=D
	for i in [1..m]:
		CM[0,i]=1
		CM[i,0]=1
	
	

	return CM,c,x,R
    
    
## 2.CM2vars

#finds the variables in every embedding equation in dim from CM graph
#requires remove intersection
#INPUT: symbolic CM matrix, variables in the matrix, coefficients of the matrix, dimension of the embedding (graph2CM can give this input easily)
#OUTPUT: variables (vars) and coefficients (c_pars) of every determinantal equation. C_sys gives the aubmatrix that specifies these equations

def CM2vars(CM,X,c,dim):
    k=dim+2
    m=CM.nrows()-1
    C=Combinations([1..m],k).list()
    l=len(C)
    vars=[]
    c_pars=[]
    C_sys=[]
   
    for i in range(l):
        CM_temp=CM[C[i],C[i]].list()
        vars_temp=uniq(remove_intersection([0,1],remove_intersection(c,CM_temp)))
	c_pars_temp=uniq(remove_intersection([0,1],remove_intersection(X,CM_temp)))
        vars.append(vars_temp)
	c_pars.append(c_pars_temp)
	C_sys.append(C[i])
    

    return vars,c_pars,C_sys


## 3.remove_intersection

#INPUT: lists a,b
#OUTPUT: b_new=b-a

def remove_intersection(a,b):
    b_new=list(copy(b))

	
    for i in range(len(b)):
	if b[i] in a:
		b_new.remove(b[i])

    

    return b_new



## 4.covering_lengths

## checks if c_pars of a given subsystem covers all edge lengths-->well constrained system
##c_pars always subsets of c
##output:1=>covering, 0=> not covering 

def covering_lengths(L_sys,c,c_pars):
    c_temp=[]
    for i in range(len(L_sys)):
	c_temp=uniq(c_temp+c_pars[L_sys[i]])
    
    
    if len(c)==len(c_temp):
	flag=1
    else:
	flag=0
   
    return flag


## 5.CM2CM_sub

#finds the variables in every embedding equation in dim from CM graph
#requires remove intersection, covering_lengths
#INPUT: X,vars,c,c_pars is obtained by CM2vars/ n is the number of variables for our square system of equations (the optimal for an embedding of a graph with m vertices in R^d is n=m-(d+1))
#OUTPUT: Well constrained subsystem(s) 

def CM2CM_sub(X,vars,c,c_pars,n):
    L_sub=[]
    C=Combinations(X,n).list()
    l=len(C)

    for i in range(l):
    	L_temp=[]
        j=0
        while j<len(vars) and len(L_temp)<n+2:
             test=remove_intersection(C[i],vars[j])

	     if len(test)==0:
		  L_temp.append(j)
             j=j+1

        if len(L_temp)==n and covering_lengths(L_temp,c,c_pars)==1:
	    L_sub.append(L_temp)
	 

    return L_sub


## 6. exp_vector

#returns the exponent vectors of a polynomial in variables X and (possibly extended with coefs) ring R
#X,f,Coef should be type 'sage.rings.polynomial...' "
#INPUT: function, variables
#OUTPUT: exponents of polynomial, variables with positive power

def exp_vector(f,X):
	
	l_x=len(X)

	Exp_vector=[]
	F=f.monomials()
	l_f=len(F)

	for j in range(l_f):
		exp_temp=[]
		for k in range(l_x):
			exp_temp.append(F[j].degree(X[k]))	
				

		if not exp_temp in Exp_vector:
			Exp_vector.append(exp_temp)

	Exp_vector_fin=[]

	for i in range(len(Exp_vector)):
		Exp_vector_fin.append(tuple(Exp_vector[i]))
	
	return Exp_vector_fin
    
## 7.graph2MV_dist

#finds MV of distance systems derived from graph of dimension dim
#INPUT: graph-G, dimension-dim, size of square subsystem-k, choose one CM submatrix-flag=0, choose all possible subsystems- flag=1
#requires to load graph2CM.sage, CM2vars.sage, CM2CM_sub.sage, covering_lengths.sage, exp_vector.sage
#needs phcpy in the path


def graph2MV_dist(G,dim,k,flag=0):

    CM,c,X,R=graph2CM(G,dim)
    vars,c_pars,C_sys=CM2vars(CM,X,c,dim)
    L_sub=CM2CM_sub(X,vars,c,c_pars,k)
    MV=[]
    
    if flag==1 and len(L_sub)>0:
        for j in range(len(L_sub)):
            L=copy(L_sub[j])
            vars_temp=[]
            CM_sys=[]
            F=[]
            for i in range(k):
                vars_temp=vars_temp+vars[L[i]]
		C_sys_temp=copy(C_sys[L[i]])
		C_sys_temp.append(0)
		CM_sys.append(CM[C_sys_temp,C_sys_temp])
		F.append(CM_sys[i].det())

	    vars_temp=uniq(vars_temp)
	    mixed_con=[p-p+1 for p in range(k)]

	    NP=[]
	    for i in range(k):
                NP_temp=exp_vector(F[i],vars_temp)
		NP.append(NP_temp)
	
            from phcpy.polytopes import mixed_volume as mv
	    MV_temp=mv(mixed_con,NP, checkin=False)
	    MV.append(MV_temp)

    elif flag==0 and len(L_sub)>0:
        L=copy(L_sub[0])
        vars_temp=[]
        CM_sys=[]
        F=[]
        for i in range(k):
            vars_temp=vars_temp+vars[L[i]]
	    C_sys_temp=copy(C_sys[L[i]])
	    C_sys_temp.append(0)
	    CM_sys.append(CM[C_sys_temp,C_sys_temp])
	    F.append(CM_sys[i].det())
        
        
        vars_temp=uniq(vars_temp)
	mixed_con=[p-p+1 for p in range(k)]

	NP=[]
	for i in range(k):
            NP_temp=exp_vector(F[i],vars_temp)
	    NP.append(NP_temp)
	
        from phcpy.polytopes import mixed_volume as mv
	MV_temp=mv(mixed_con,NP)
	MV.append(MV_temp)
        
   
    

    return MV



    
