######## Includes all functions used to compute the embeddings of minimally rigid graphs in R^3 using sphere equations

## 1.graph_triangles

#INPUT: graph
#OUTPUT: list of pseudo-triangles

def graph_triangles(G):
	D=G.adjacency_matrix()
	
	n=G.order()

	T=[]
	for i in range(n-2):
		for j in range(i+1,n-1):
			if D[i,j]==1:
				for k in range(j+1,n):
					if D[i,k]==1 and D[j,k]==1:
						T.append([i,j,k])


	return T



## 2.remove_intersection

#INPUT: lists a,b
#OUTPUT: b_new=b-a

def remove_intersection(a,b):
    b_new=list(copy(b))

	
    for i in range(len(b)):
	if b[i] in a:
		b_new.remove(b[i])

    

    return b_new


## 3.adjacent2first_tr

#finds a permutation of adjacent matrix such that the 3 first vertices form a triangle with maximum degree for them"
#dependencies:graph_triangles, remove_intersection"

def adjacent2first_tr(G):
    
    T=graph_triangles(G)
    A=G.adjacency_matrix()
    n=A.nrows()
    p=range(n) 
    Deg=G.degree()

    if len(T)>0:
        Deg_tr=[]
        for i in range(len(T)):
	    Deg_tr.append(Deg[T[i][1]]+Deg[T[i][2]]+Deg[T[i][0]])
        L=list(enumerate(Deg_tr, 0))
        Ind=L[-1][0]

	p_tr=T[Ind]
	p_ter=remove_intersection(p_tr,p)
	p=p_tr+p_ter
        B=A.matrix_from_rows_and_columns(p,p)
    else:
	B=[]
	

    

    
    

    return B


## 4.all_adjacent_tr
 


#finds all permutations of adjacent matrix such that the 3 first vertices form a triangle with maximum degree for them
#dependencies:graph_triangles, remove_intersection

def all_adjacent_tr(G):
    
    T=graph_triangles(G)
    A=G.adjacency_matrix()
    n=A.nrows()
    p=range(n) 
    Deg=G.degree()
    B_all=[]


    
    if len(T)>0:
	for i in range(len(T)):
        	p_tr=T[i]
		p_ter=remove_intersection(p_tr,p)
		p_new=p_tr+p_ter
        	B=A.matrix_from_rows_and_columns(p_new,p_new)
		B_all.append(B)
    
	

    

    return B_all




## 5.dis_sq

#distance squared matrix of a conformation in nxdim


def dis_sq(A,R=RR):

    n=A.nrows()
    d=A.ncols()
    Dis=Matrix(R,n,n)    
    
    
    for i in range(n-1):
	for j in range(i+1,n):
		Dis[i,j]=(norm(A[i,:]-A[j,:]))^2
		Dis[j,i]=Dis[i,j]
		

	

   
    return Dis


## 6.graph2MV_sphere

#requires phcpy in the path of sage


def graph2MV_sphere(G):
    from phcpy.polytopes import mixed_volume as mv

    n=G.order()
    B=adjacent2first_tr(G)

    "points"
    P=[]

    
    vars=4*(n-3)

    "magnitudes"
    for i in range(n-3):
        x_temp=[0*p for p in range(vars)]
        y_temp=[0*p for p in range(vars)]
        z_temp=[0*p for p in range(vars)]
        s_temp=[0*p for p in range(vars)]
        x_temp[4*i]=2
        y_temp[4*i+1]=2
        z_temp[4*i+2]=2
        s_temp[4*i+3]=1
 
        p_temp=[tuple(x_temp),tuple(y_temp),tuple(z_temp),tuple(s_temp)]
        P.append(p_temp)
    
    "equations with vertex 1"
    for i in range(n-3):
	if B[0,i+3]==1:
	    
            s_temp=[0*p for p in range(vars)]
	    coeff=[0*p for p in range(vars)]
            s_temp[4*i+3]=1
 
            p_temp=[tuple(s_temp),tuple(coeff)]
            P.append(p_temp)
            
    "equations with vertex 2"
    for i in range(n-3):
	if B[1,i+3]==1:
	    y_temp=[0*p for p in range(vars)]
            s_temp=[0*p for p in range(vars)]
	    coeff=[0*p for p in range(vars)]

            s_temp[4*i+3]=1
            y_temp[4*i+1]=1
 
            p_temp=[tuple(y_temp),tuple(s_temp),tuple(coeff)]
            P.append(p_temp)

    "equations with vertex 3"
    for i in range(n-3):
	if B[2,i+3]==1:
	    
            z_temp=[0*p for p in range(vars)]
            s_temp=[0*p for p in range(vars)]
	    coeff=[0*p for p in range(vars)]

            s_temp[4*i+3]=1
            z_temp[4*i+2]=1
 
            p_temp=[tuple(z_temp),tuple(s_temp),tuple(coeff)]
            P.append(p_temp)


   
    "other"
    for i in range(n-4):
        for j in range(i+1,n-3):
	    if B[i+3,j+3]==1:

                x_temp=[0*p for p in range(vars)]
	        y_temp=[0*p for p in range(vars)]
                z_temp=[0*p for p in range(vars)]
                s1_temp=[0*p for p in range(vars)]
	        s2_temp=[0*p for p in range(vars)]
	        coeff=[0*p for p in range(vars)]

                x_temp[4*i]=1
                x_temp[4*j]=1
                y_temp[4*i+1]=1
                y_temp[4*j+1]=1
                z_temp[4*i+2]=1
                z_temp[4*j+2]=1

                s1_temp[4*i+3]=1
                s2_temp[4*j+3]=1

                p_temp=[tuple(x_temp),tuple(y_temp),tuple(z_temp),tuple(s1_temp),tuple(s2_temp),tuple(coeff)]
                P.append(p_temp)



    


    mixed_con=[p-p+1 for p in range(vars)]
    MV=mv(mixed_con,P)
    
    


    return MV



## 7.graph2MV_sphere_adj


## Finds mixed volume of a minimally rigid graph in R^3 represented by an adjacent matrix that specifies the initial triangle
## phcpy should be added in sage's path

def graph2MV_sphere_adj(B):
    from phcpy.polytopes import mixed_volume as mv

    n=B.nrows()
    
    "points"
    P=[]

    
    vars=4*(n-3)

    "magnitudes"
    for i in range(n-3):
        x_temp=[0*p for p in range(vars)]
        y_temp=[0*p for p in range(vars)]
        z_temp=[0*p for p in range(vars)]
        s_temp=[0*p for p in range(vars)]
        x_temp[4*i]=2
        y_temp[4*i+1]=2
        z_temp[4*i+2]=2
        s_temp[4*i+3]=1
 
        p_temp=[tuple(x_temp),tuple(y_temp),tuple(z_temp),tuple(s_temp)]
        P.append(p_temp)
    
    "equations with vertex 1"
    for i in range(n-3):
	if B[0,i+3]==1:
	    
            s_temp=[0*p for p in range(vars)]
	    coeff=[0*p for p in range(vars)]
            s_temp[4*i+3]=1
 
            p_temp=[tuple(s_temp),tuple(coeff)]
            P.append(p_temp)
            
    "equations with vertex 2"
    for i in range(n-3):
	if B[1,i+3]==1:
	    y_temp=[0*p for p in range(vars)]
            s_temp=[0*p for p in range(vars)]
	    coeff=[0*p for p in range(vars)]

            s_temp[4*i+3]=1
            y_temp[4*i+1]=1
 
            p_temp=[tuple(y_temp),tuple(s_temp),tuple(coeff)]
            P.append(p_temp)

    "equations with vertex 3"
    for i in range(n-3):
	if B[2,i+3]==1:
	    
            z_temp=[0*p for p in range(vars)]
            s_temp=[0*p for p in range(vars)]
	    coeff=[0*p for p in range(vars)]

            s_temp[4*i+3]=1
            z_temp[4*i+2]=1
 
            p_temp=[tuple(z_temp),tuple(s_temp),tuple(coeff)]
            P.append(p_temp)


   
    "other"
    for i in range(n-4):
        for j in range(i+1,n-3):
	    if B[i+3,j+3]==1:

                x_temp=[0*p for p in range(vars)]
	        y_temp=[0*p for p in range(vars)]
                z_temp=[0*p for p in range(vars)]
                s1_temp=[0*p for p in range(vars)]
	        s2_temp=[0*p for p in range(vars)]
	        coeff=[0*p for p in range(vars)]

                x_temp[4*i]=1
                x_temp[4*j]=1
                y_temp[4*i+1]=1
                y_temp[4*j+1]=1
                z_temp[4*i+2]=1
                z_temp[4*j+2]=1

                s1_temp[4*i+3]=1
                s2_temp[4*j+3]=1

                p_temp=[tuple(x_temp),tuple(y_temp),tuple(z_temp),tuple(s1_temp),tuple(s2_temp),tuple(coeff)]
                P.append(p_temp)



    


    mixed_con=[p-p+1 for p in range(vars)]
    MV=mv(mixed_con,P)
    
    


    return MV


## 8.sols_adj_phc

#returns solutions of naive system given an adjacent matrix and its dimension
#tracks previous solution if given as flag_prev=[sol_prev, sys_prev], else flag_prev=0
#the system needs to be square
#dimension 3


def sols_adj_phc(B,Dis,flag_prev=0):

    F=[]
    a=sqrt(Dis[0,1])
    b=(Dis[0,2]+Dis[0,1]-Dis[1,2])/(2*a)
    c=sqrt(Dis[0,2]-b^2)
    fixed=[float(a),float(b),float(c)]

    n=B.nrows() 
    x=list(var('x_%d' % (i+1)) for i in range(3,n))
    y=list(var('y_%d' % (i+1)) for i in range(3,n))
    z=list(var('z_%d' % (i+1)) for i in range(3,n))
    s=list(var('s_%d' % (i+1)) for i in range(3,n))
    vars=x+y+z+s

    #magnitude equations
    for i in range(3,n):
	f_temp=s[i-3]-x[i-3]^2-y[i-3]^2-z[i-3]^2
	F.append(f_temp)

    for i in range(3,n):
	if B[0,i]==1:
		f_temp=s[i-3]-Dis[0,i]
		F.append(f_temp)

    
		
    for i in range(3,n):
	if B[1,i]==1:
		f_temp=s[i-3]+Dis[0,1]-2*y[i-3]*fixed[0]-Dis[1,i]
		F.append(f_temp)
    

    for i in range(3,n):
	if B[2,i]==1:
		f_temp=s[i-3]+Dis[0,2]-2*y[i-3]*fixed[1]-2*z[i-3]*fixed[2]-Dis[2,i]
		F.append(f_temp)


    for i in range(3,n-1):
	for j in range(i+1,n):
		if B[i,j]==1:
			f_temp=s[i-3]+s[j-3]-2*x[i-3]*x[j-3]-2*y[i-3]*y[j-3]-2*z[i-3]*z[j-3]-Dis[i,j]
			F.append(f_temp)


    from phcpy.solver import solve
    from phcpy.trackers import track


    f=[]
    for i in range(len(F)):
	f_temp=str(F[i])
	f_temp+=';'
	f.append(f_temp)

    if flag_prev==0:
	sols=solve(f, verbose=False, tasks=2)
    else:
	sols_prev=flag_prev[0]
	f_prev=flag_prev[1]
	sols=track(f,f_prev,sols_prev, tasks=2)

    return sols, f


## 9.int2graph

# INPUT: an integer
#OUTPUT: the corresponding graph

def int2graph(N):
    L=Integer(N).binary()
    n=ceil((1+sqrt(1+8*len(L)))/2)
    #print N, (L), n
    adjacencyMatrix=matrix(n,n)
    j=len(L)
    r=n-1
    c=n-1
    row_len=n-1
    while j>0:
        #print adjacencyMatrix[r,c]
        c+=-1
        if c<=row_len:
            row_len+=-1
            c=n-1
            r+=-1
        j-=1
        adjacencyMatrix[r,c]=L[j]

    return Graph(adjacencyMatrix+adjacencyMatrix.transpose())




## 10.count_sols_sphere_phc



#Input: list of integer represantations of graphs/same vertices (minimally rigid graphs in dimension 3)
#Output: list with the number of generic complex solutions for each graph
#needs to load functions dis_sq, adjacent2first_tr, graph_triangles, remove_intersection, sols_adj_phc
#phcpy should be added in sage's path

def count_sols_sphere_phc(g):
    d=int2graph(g[0]).order()
    X= random_matrix(RR,d,3)
    X[0,0]=0
    X[0,1]=0
    X[0,2]=0
    X[1,0]=0
    X[1,2]=0
    X[2,0]=0
    Dis=dis_sq(X)
    sols_count=[]

    for i in range(len(g)):
	B=adjacent2first_tr(int2graph(g[i]))
	sols_temp, f=sols_adj_phc(B,Dis,)
        n=len(sols_temp)
	sols_count.append(n)

    return sols_count




